"""
.. py:module:: medic.orm
   :synopsis: A SQLAlchemy-based MEDLINE DB ORM.

.. moduleauthor:: Florian Leitner <florian.leitner@gmail.com>
.. License: GNU Affero GPL v3 (http://www.gnu.org/licenses/agpl.html)
"""

import logging
from datetime import date
from sqlalchemy import engine, select, and_, Enum
from sqlalchemy import event
# from sqlalchemy.engine import RowProxy
from sqlalchemy.engine.url import URL
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relation, session
from sqlalchemy.orm.collections import column_mapped_collection
from sqlalchemy.schema import \
    Column, CheckConstraint, ForeignKeyConstraint, ForeignKey, Index
from sqlalchemy.types import \
    Boolean, BigInteger, Date, SmallInteger, Unicode, UnicodeText

__all__ = [
    'Citation', 'Abstract', 'Author', 'Chemical', 'Database', 'Descriptor',
    'Identifier', 'Keyword', 'PublicationType', 'Qualifier', 'Section'
]

_Base = declarative_base()
_db = None
_session = lambda *args, **kwds: None

NULL = lambda s: '\\N' if s is None else s
DATE = lambda s: '\\N' if s is None else s.isoformat()
STRING = lambda s: s.replace('\\', '\\\\').replace('\t', '\\t').replace('\n', '\\n')

logger = logging.getLogger(__name__)


def InitDb(*args, **kwds):
    """
    Create a new DBAPI connection pool.

    The most common and only required argument is the connection URL.
    The URL can either be a string or a `sqlalchemy.engine.url.URL`.
    This method has not return value and needs to be called only once per process.

    See `sqlalchemy.engine.create_engine`.
    """
    global _Base
    global _db
    global _session

    if len(args) > 0:
        # inject the foreign key pragma when using SQLite databases to ensure integrity
        # http://docs.sqlalchemy.org/en/rel_0_8/dialects/sqlite.html#foreign-key-support
        if (isinstance(args[0], str) and args[0].startswith('sqlite')) or \
                (isinstance(args[0], URL) and args[0].get_dialect() == 'sqlite'):
            # noinspection PyUnusedLocal
            @event.listens_for(engine.Engine, "connect")
            def set_sqlite_pragma(dbapi_connection, _):
                """Injection to enable foreign keys and make SQLite a bit faster."""
                cursor = dbapi_connection.cursor()
                cursor.execute("PRAGMA foreign_keys=ON")
                cursor.execute("PRAGMA synchronous=OFF")
                cursor.execute("PRAGMA page_size=65536")
                cursor.close()

    _db = engine.create_engine(*args, **kwds)
    _Base.metadata.create_all(_db)
    _session = session.sessionmaker(bind=_db)
    logger.debug("DB bound to %s", _db)
    return None


def Session(*args, **kwds):
    """
    Create a new DBAPI session object to work with.

    See `sqlalchemy.orm.session.Session`.
    """
    return _session(*args, **kwds)


def _fetch_first(query):
    """Given a *query*, fetch the first row and return the first element or ``None``."""
    conn = _db.engine.connect()
    logger.debug("%s", query)

    try:
        result = conn.execute(query)
        row = result.first()
        return row[0] if row else None
    finally:
        conn.close()


def _fetch_all(query):
    """Given a *query*, fetch and return all rows."""
    conn = _db.engine.connect()
    logger.debug("%s", query)

    try:
        result = conn.execute(query)
        return result.fetchall()
    finally:
        conn.close()


# noinspection PyUnresolvedReferences
class SelectMixin(object):
    """
    Mixin for child tables to select rows (and columns)
    for a particular parent object
    (i.e., `Citation`, except for `Qualifier`).
    """

    @classmethod
    def _buildQuery(cls, columns, parent_pk):
        if isinstance(parent_pk, tuple):
            last = len(cls.__table__.primary_key) - 1
            assert len(parent_pk) == last, \
                "wrong parent_pk %s (%d)" % (str(parent_pk), len(cls.__table__.primary_key))
            where_clause = [
                col == parent_pk[idx] for idx, col in
                enumerate(cls.__table__.primary_key) if idx != last
            ]
            return select(columns, and_(*where_clause))
        else:
            assert isinstance(parent_pk, int), \
                "parent_pk not a PMID: %s" % str(parent_pk)
            return select(columns, cls.__table__.c.pmid == parent_pk)

    @classmethod
    def select(cls, parent_pk, attributes):
        """
        Return *attributes* columns for all rows that match the object's
        *parent_pk* (the pmid in most cases, but (pmid, num) for
        `Qualifier`).

        :arg  parent_pk: The parent objects pk
        :type parent_pk: :class:`tuple` or :keyword:`scalar`

        :arg  attributes: A list of attribute (column) names
        :type attributes: :class:`list` of :class:`str`
        """
        mapping = dict((col.key, col) for col in cls.__table__.c)
        columns = [mapping[name] for name in attributes]
        query = cls._buildQuery(columns, parent_pk)
        return _fetch_all(query)

    @classmethod
    def selectAll(cls, parent_pk):
        """
        Return all columns for all rows that match the *parent_pk*.

        See :meth:`select` for details.
        """
        query = cls._buildQuery([cls.__table__], parent_pk)
        return _fetch_all(query)


class Identifier(_Base, SelectMixin):
    """
    All known unique IDs for a PubMed record.

    Attributes:

        pmid
            the record this alternate ID belongs to
        namespace
            the type of ID (doi, pii, pmc, pubmed, etc.)
        value
            the actual ID string

    Primary Key: ``(pmid, namespace)``
    """

    __tablename__ = 'identifiers'

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    namespace = Column(
        Unicode(length=32), CheckConstraint("namespace <> ''"), primary_key=True
    )
    value = Column(
        Unicode(length=256), CheckConstraint("value <> ''"), nullable=False
    )

    def __init__(self, pmid: int, namespace: str, value: str):
        assert pmid > 0, pmid
        assert namespace
        assert value
        self.pmid = pmid
        self.namespace = namespace
        self.value = value

    def __str__(self):
        return '{}\t{}\t{}\n'.format(
            NULL(self.pmid), NULL(self.namespace), STRING(self.value)
        )

    def __repr__(self):
        return "Identifier<{}:{}>".format(self.pmid, self.namespace)

    def __eq__(self, other):
        return isinstance(other, Identifier) and \
            self.pmid == other.pmid and \
            self.namespace == other.namespace and \
            self.value == other.value

    @classmethod
    def pmid2doi(cls, pmid: int):
        """Convert a PMID to a DOI (or ``None`` if no mapping is found)."""
        c = cls.__table__.c
        query = select(
            [c.value], (c.namespace == 'doi') & (c.pmid == pmid)
        )
        return _fetch_first(query)

    @classmethod
    def doi2pmid(cls, doi: str):
        """Convert a DOI to a PMID (or ``None`` if no mapping is found)."""
        c = cls.__table__.c
        query = select(
            [c.pmid], (c.namespace == 'doi') & (c.value == doi)
        )
        return _fetch_first(query)

    @classmethod
    def mapDois2Pmids(cls, dois: list):
        """
        Return a mapping :class:`dict` for a list of DOIs to their PMIDs
        (or and empty :class:`dict` if no mapping is found).

        If for a given DOI no mapping exists, it is no included in the
        returned dictionary.
        """
        if not len(dois):
            return {}

        c = cls.__table__.c
        query = select(
            [c.value, c.pmid], (c.namespace == 'doi') & c.value.in_(dois)
        )
        mappings = _fetch_all(query)
        return dict(mappings) if mappings is not None else {}

    @classmethod
    def mapPmids2Dois(cls, pmids: list):
        """
        Return a mapping :class:`dict` for a list of PMIDs to their DOIs
        (or and empty :class:`dict` if no mapping is found).

        If for a given PMID no mapping exists, it is no included in the
        returned dictionary.
        """
        if not len(pmids):
            return {}

        t = cls.__table__
        query = select(
            [t.c.pmid, t.c.value], (t.c.namespace == 'doi') & t.c.pmid.in_(pmids)
        )
        mappings = _fetch_all(query)
        return dict(mappings) if mappings is not None else {}


# Index to make queries for a particular ID, e.g., a DOI, faster.
Index(
    'identifiers_namespace_value_idx',
    Identifier.__table__.c.namespace, Identifier.__table__.c.value
)


class Author(_Base, SelectMixin):
    """
    Author names for a PubMed record.

    Attributes:

        pmid
            the record this author belongs to
        pos
            the order/position of the name in the PubMed record (starting from 1)
        name
            an author's last name or the collective's name (never empty)
        initials
            an author's initials *
        forename
            the expansion of the initials known to PubMed *
        suffix
            an author name's suffix *

    * empty string if explicitly non-existent, NULL if unknown

    Primary Key: ``(pmid, pos)``
    """

    __tablename__ = 'authors'

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    pos = Column(SmallInteger, CheckConstraint("pos > 0"), primary_key=True)
    name = Column(
        UnicodeText, CheckConstraint("name <> ''"), nullable=False
    )
    initials = Column(Unicode(length=128), nullable=True)
    forename = Column(Unicode(length=256), nullable=True)
    suffix = Column(Unicode(length=128), nullable=True)

    def __init__(self, pmid: int, pos: int, name: str,
                 initials: str=None, forename: str=None, suffix: str=None):
        assert pmid > 0, pmid
        assert pos > 0, pos
        assert name, repr(name)
        self.pmid = pmid
        self.pos = pos
        self.name = name
        self.initials = initials
        self.forename = forename
        self.suffix = suffix

    def __str__(self):
        return "{}\t{}\t{}\t{}\t{}\t{}\n".format(
            NULL(self.pmid), NULL(self.pos), STRING(self.name),
            NULL(self.initials), NULL(self.forename), NULL(self.suffix))

    def __repr__(self):
        return "Author<{}:{}>".format(self.pmid, self.pos)

    def __eq__(self, other):
        return isinstance(other, Author) and \
            self.pmid == other.pmid and \
            self.pos == other.pos and \
            self.name == other.name and \
            self.initials == other.initials and \
            self.forename == other.forename and \
            self.suffix == other.suffix

    def fullName(self):
        """
        Return the full name of this author
        (using forename or initials, last, and suffix).
        """
        name = []
        if self.forename:
            name.append(self.forename)
        elif self.initials:
            name.append(self.initials)
        name.append(self.name)
        if self.suffix:
            name.append(self.suffix)
        return ' '.join(name)

    def shortName(self):
        """Return the short name of this author (using initials and last)."""
        name = []
        if self.initials:
            name.append(self.initials)
        elif self.forename:
            name.append(''.join([n[0] for n in self.forename.split()]))
        name.append(self.name)
        return ' '.join(name)


class Qualifier(_Base, SelectMixin):
    """
    One of a MeSH descriptor's qualifiers for a record.

    Attributes:

        pmid
            the record this qualifier name belongs to
        num
            the descriptor order in the record (starting from 1)
        sub
            the qualifier order within the descriptor (starting from 1)
        major
            ``True`` if major, ``False`` if minor
        name
            the qualifier (name)

    Primary Key: ``(pmid, num, sub)``
    """

    __tablename__ = 'qualifiers'
    __table_args__ = (
        ForeignKeyConstraint(
            ('pmid', 'num'), ('descriptors.pmid', 'descriptors.num'),
            ondelete="CASCADE"
        ),
    )

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    num = Column(SmallInteger, primary_key=True)
    sub = Column(SmallInteger, CheckConstraint("sub > 0"), primary_key=True)
    major = Column(Boolean, nullable=False)
    name = Column(UnicodeText, CheckConstraint("name <> ''"), nullable=False)

    def __init__(self, pmid: int, num: int, sub: int, name: str, major: bool=False):
        assert pmid > 0, pmid
        assert num > 0, num
        assert sub > 0, sub
        assert name, repr(name)
        self.pmid = pmid
        self.num = num
        self.sub = sub
        self.major = major
        self.name = name

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\n'.format(
            NULL(self.pmid), NULL(self.num), NULL(self.sub),
            'T' if self.major else 'F', STRING(self.name),
        )

    def __repr__(self):
        return "Qualifier<{}:{}:{}>".format(self.pmid, self.num, self.sub)

    def __eq__(self, other):
        return isinstance(other, Qualifier) and \
            self.pmid == other.pmid and \
            self.num == other.num and \
            self.sub == other.sub and \
            self.name == other.name and \
            self.major == other.major


class Descriptor(_Base, SelectMixin):
    """
    A MeSH descriptor for the record.

    Attributes:

        pmid
            the record this descriptor name belongs to
        num
            the descriptor order in the record (starting from 1)
        major
            ``True`` if major, ``False`` if minor
        name
            the descriptor

    Relations:

        qualifiers
            a :class:`list` of the descriptor's qualifiers

    Primary Key: ``(pmid, num)``
    """

    __tablename__ = 'descriptors'

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    num = Column(SmallInteger, CheckConstraint("num > 0"), primary_key=True)
    major = Column(Boolean, nullable=False)
    name = Column(UnicodeText, CheckConstraint("name <> ''"), nullable=False)

    qualifiers = relation(
        Qualifier, backref='descriptor', cascade="all",
        order_by=Qualifier.__table__.c.sub
    )

    def __init__(self, pmid: int, num: int, name: str, major: bool=False):
        assert pmid > 0, pmid
        assert num > 0, num
        assert name, repr(name)
        self.pmid = pmid
        self.num = num
        self.major = major
        self.name = name

    def __str__(self):
        return '{}\t{}\t{}\t{}\n'.format(
            NULL(self.pmid), NULL(self.num), 'T' if self.major else 'F', STRING(self.name)
        )

    def __repr__(self):
        return "Descriptor<{}:{}>".format(self.pmid, self.num)

    def __eq__(self, other):
        return isinstance(other, Descriptor) and \
            self.pmid == other.pmid and \
            self.num == other.num and \
            self.name == other.name and \
            self.major == other.major


class Chemical(_Base, SelectMixin):
    """
    References to chemicals and substances curated my the NLM.

    Attributes:

        pmid
            the record's identifier (PubMed ID)
        idx
            the order in the XML record
        uid
            the chemical's "unique identifier" (registry number, etc.)
        name
            the chemical's name

    Primary Key: ``(pmid, idx)``
    """

    __tablename__ = 'chemicals'

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    idx = Column(SmallInteger, CheckConstraint("idx > 0"), primary_key=True)
    uid = Column(Unicode(length=256), CheckConstraint("uid <> ''"), nullable=True)
    name = Column(Unicode(length=256), CheckConstraint("name <> ''"), nullable=False)

    def __init__(self, pmid: int, idx: int, name: str, uid: str=None):
        assert pmid > 0, pmid
        assert idx > 0, idx
        assert name, repr(name)
        assert uid is None or uid, repr(uid)
        self.pmid = pmid
        self.idx = idx
        self.uid = uid
        self.name = name

    def __str__(self):
        return '{}\t{}\t{}\t{}\n'.format(
            NULL(self.pmid), NULL(self.idx), NULL(self.uid), STRING(self.name)
        )

    def __repr__(self):
        return "Chemical<{}:{}>".format(self.pmid, self.idx)

    def __eq__(self, other):
        return isinstance(other, Chemical) and \
            self.pmid == other.pmid and \
            self.idx == other.idx and \
            self.uid == other.uid and \
            self.name == other.name


class PublicationType(_Base, SelectMixin):
    """
    Records the type of publication this citation refers to.

    Attributes:

        pmid
            the record's identifier (PubMed ID)
        value
            the publication type value

    Primary Key: ``(pmid, value)``
    """

    __tablename__ = 'publication_types'

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    value = Column(Unicode(length=256), CheckConstraint("value <> ''"), primary_key=True)

    def __init__(self, pmid: int, value: str):
        assert pmid > 0, pmid
        assert value, repr(value)
        self.pmid = pmid
        self.value = value

    def __str__(self):
        return '{}\t{}\n'.format(NULL(self.pmid), NULL(self.value))

    def __repr__(self):
        return "PublicationType<{}:{}>".format(self.pmid, self.value)

    def __eq__(self, other):
        return isinstance(other, PublicationType) and \
            self.pmid == other.pmid and \
            self.value == other.value


class Database(_Base, SelectMixin):
    """
    References to external databases curated by the NLM.

    Attributes:

        pmid
            the record's identifier (PubMed ID)
        name
            the referenced database' name
        accession
            the referenced record's identifier

    Primary Key: ``(pmid, name, accession)``
    """

    __tablename__ = 'databases'

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    name = Column(Unicode(length=256), CheckConstraint("name <> ''"), primary_key=True)
    accession = Column(Unicode(length=256), CheckConstraint("accession <> ''"), primary_key=True)

    def __init__(self, pmid: int, name: str, accession: str):
        assert pmid > 0, pmid
        assert name, repr(name)
        assert accession, repr(accession)
        self.pmid = pmid
        self.name = name
        self.accession = accession

    def __str__(self):
        return '{}\t{}\t{}\n'.format(
            NULL(self.pmid), NULL(self.name), STRING(self.accession)
        )

    def __repr__(self):
        return "Database<{}:{}:{}>".format(self.pmid, self.name, self.accession)

    def __eq__(self, other):
        return isinstance(other, Database) and \
            self.pmid == other.pmid and \
            self.name == other.name and \
            self.accession == other.accession


class Keyword(_Base, SelectMixin):
    """
    Keywords, external or curated by the NLM.

    Attributes:

        pmid
            the record's identifier (PubMed ID)
        owner
            the entity that provided the keyword
        cnt
            a unique counter for all keywords from a given owner and record
            (starting from 1)
        major
            if the keyword is a major topic of this article
        name
            the keyword itself

    Primary Key: ``(pmid, owner, cnt)``
    """

    __tablename__ = 'keywords'

    OWNERS = frozenset({'NASA', 'PIP', 'KIE', 'NLM', 'NOTNLM', 'HHS'})

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    owner = Column(Enum(*OWNERS, name='owner'), primary_key=True)
    cnt = Column(SmallInteger, CheckConstraint("cnt > 0"), primary_key=True)
    major = Column(Boolean, nullable=False)
    name = Column(UnicodeText, CheckConstraint("name <> ''"), nullable=False)

    def __init__(self, pmid: int, owner: str, cnt: int, name: str, major: bool=False):
        assert pmid > 0, pmid
        assert owner in Keyword.OWNERS, repr(owner)
        assert cnt > 0, cnt
        assert name, repr(name)
        self.pmid = pmid
        self.owner = owner
        self.cnt = cnt
        self.major = major
        self.name = name

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\n'.format(
            NULL(self.pmid), NULL(self.owner), NULL(self.cnt),
            'T' if self.major else 'F', STRING(self.name)
        )

    def __repr__(self):
        return "Keyword<{}:{}:{}>".format(
            self.pmid, self.owner, self.cnt
        )

    def __eq__(self, other):
        return isinstance(other, Keyword) and \
            self.pmid == other.pmid and \
            self.owner == other.owner and \
            self.cnt == other.cnt and \
            self.major == other.major and \
            self.name == other.name


class Section(_Base, SelectMixin):
    """
    The text sections of the records.

    Attributes:

        pmid
            the record's identifier (PubMed ID)
        source
            the abstract's source (see `Abstract.SOURCES`)
        seq
            the sequence of sections in the abstract (starting from 1)
        name
            the name of the section (Abstract, Background, Methods, Unassigned, ...)
        label
            section label as defined by the publisher (if any)
        content
            the text content of the section
        truncated
            if the text content was truncated by PubMed
            that means the content contained the string "(ABSTRACT TRUNCATED AT 250 WORDS)" at
            its end; this message is removed by the parser

    Primary Key: ``(pmid, source, seq)``
    """
    __tablename__ = 'sections'
    __table_args__ = (
        ForeignKeyConstraint(
            ('pmid', 'source'), ('abstracts.pmid', 'abstracts.source'),
            ondelete="CASCADE"
        ),
    )

    SOURCES = frozenset({'NLM', 'AAMC', 'AIDS', 'KIE', 'PIP', 'NASA', 'Publisher'})

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    source = Column(Enum(*SOURCES, name='source'), primary_key=True)
    seq = Column(SmallInteger, CheckConstraint("seq > 0"), primary_key=True)
    name = Column(Unicode(length=64), CheckConstraint("name <> ''"), nullable=False)
    label = Column(Unicode(length=256), CheckConstraint("label <> ''"), nullable=True)
    content = Column(UnicodeText, CheckConstraint("content <> ''"), nullable=False)
    truncated = Column(Boolean, nullable=False, default=False)

    def __init__(self, pmid: int, source: str, seq: int, name: str, content: str,
                 label: str=None, truncated: bool=False):
        assert pmid > 0, pmid
        assert source in Section.SOURCES, repr(source)
        assert seq > 0, seq
        assert name, repr(name)
        assert content, repr(content)
        assert label is None or label, repr(label)
        self.pmid = pmid
        self.source = source
        self.seq = seq
        self.name = name
        self.label = label
        self.content = content
        self.truncated = bool(truncated)

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(
            NULL(self.pmid), NULL(self.source), NULL(self.seq), NULL(self.name),
            NULL(self.label), STRING(self.content), 'T' if self.truncated else 'F'
        )

    def __repr__(self):
        return "Section<{}:{}:{}>".format(self.pmid, self.source, self.seq)

    def __eq__(self, other):
        return isinstance(other, Section) and \
            self.pmid == other.pmid and \
            self.source == other.source and \
            self.seq == other.seq and \
            self.name == other.name and \
            self.label == other.label and \
            self.content == other.content and \
            self.truncated == other.truncated


class Abstract(_Base, SelectMixin):
    """
    The abstract of the records.

    Attributes:

        pmid
            the record's identifier (PubMed ID)
        source
            the abstract's source (see `Abstract.SOURCES`)
        copyright
            the abstract's copyright notice (if any)

    Primary Key: ``(pmid, source)``
    """
    __tablename__ = 'abstracts'

    SOURCES = Section.SOURCES

    pmid = Column(BigInteger, ForeignKey('citations.pmid', ondelete="CASCADE"), primary_key=True)
    source = Column(Enum(*SOURCES, name='source'), primary_key=True)
    copyright = Column(UnicodeText, CheckConstraint("copyright <> ''"), nullable=True)

    sections = relation(
        Section, backref='abstract', cascade="all",
        order_by=Section.__table__.c.seq
    )

    def __init__(self, pmid: int, source: str='NLM', copy: str=None):
        assert pmid > 0, pmid
        assert source in Abstract.SOURCES
        assert copy is None or copy, repr(copy)
        self.pmid = pmid
        self.source = source
        self.copyright = copy

    def __str__(self):
        return '{}\t{}\t{}\n'.format(
            NULL(self.pmid), NULL(self.source), NULL(self.copyright)
        )

    def __repr__(self):
        return "Abstract<{}:{}>".format(self.pmid, self.source)

    def __eq__(self, other):
        return isinstance(other, Abstract) and \
            self.pmid == other.pmid and \
            self.source == other.source and \
            self.copyright == other.copyright


class Citation(_Base):
    """
    A MEDLINE or PubMed citation record.

    Attributes:

        pmid
            the record's identifier (PubMed ID)
        status
            the current status of this record (see `Citation.STATES`)
        title
            the record's title
        journal
            the journal name (Medline TA)
        created
            the record's creation date
        completed
            the record's completion date
        revised
            the record's revision date
        modified
            the date the record was last modified in the DB

    Relations:

        sections
            a :class:`list` of the record's text sections
        authors
            a :class:`list` of the record's author names
        identifiers
            a :class:`dict` of the record's alternate IDs using the
            :attr:`AlternateID.key` a dictionary keys
        descriptors
            a :class:`list` of the record's MeSH descriptors
        qualifiers
            a :class:`list` of the record's MeSH qualifiers
        chemicals
            a :class:`list` of the record's chemicals
        databases
            a :class:`list` of the record's external DB references

    Primary Key: ``pmid``
    """

    STATES = frozenset({'Completed', 'In-Process', 'PubMed-not-MEDLINE',
                        'In-Data-Review', 'Publisher', 'MEDLINE', 'OLDMEDLINE'})
    CHILDREN = (
        Section, Identifier, Database, Author, Descriptor, Qualifier,  # Qualifier last!
    )
    TABLENAMES = [cls.__tablename__ for cls in CHILDREN]
    TABLES = {cls.__tablename__: cls.__table__ for cls in CHILDREN}

    __tablename__ = 'citations'

    abstracts = relation(
        Abstract, backref='citation', cascade='all, delete-orphan',
        collection_class=column_mapped_collection(Abstract.source)
    )
    authors = relation(
        Author, backref='citation', cascade='all, delete-orphan',
        order_by=Author.__table__.c.pos
    )
    chemicals = relation(Chemical, backref='citation', cascade='all, delete-orphan')
    databases = relation(Database, backref='citation', cascade='all, delete-orphan')
    descriptors = relation(
        Descriptor, backref='citation', cascade='all, delete-orphan',
        order_by=Descriptor.__table__.c.num
    )
    identifiers = relation(
        Identifier, backref='citation', cascade='all, delete-orphan',
        collection_class=column_mapped_collection(Identifier.namespace)
    )
    keywords = relation(
        Keyword, backref='citation', cascade='all, delete-orphan',
        order_by=Keyword.__table__.c.owner
    )
    publication_types = relation(PublicationType, backref='citation', cascade='all, delete-orphan')
    qualifiers = relation(Qualifier, backref='citation')
    sections = relation(Section, backref='citation')

    pmid = Column(BigInteger, CheckConstraint('pmid > 0'), primary_key=True, autoincrement=False)
    status = Column(Enum(*STATES, name='state'), nullable=False)
    title = Column(UnicodeText, CheckConstraint("title <> ''"), nullable=False)
    journal = Column(Unicode(length=256), CheckConstraint("journal <> ''"), nullable=False)
    pub_date = Column(Unicode(length=256), CheckConstraint("pub_date <> ''"), nullable=False)
    issue = Column(Unicode(length=256), CheckConstraint("issue <> ''"), nullable=True)
    pagination = Column(Unicode(length=256), CheckConstraint("pagination <> ''"), nullable=True)
    created = Column(Date, nullable=False)
    completed = Column(Date, nullable=True)
    revised = Column(Date, nullable=True)
    modified = Column(Date, default=date.today, onupdate=date.today, nullable=False)

    def __init__(self,
                 pmid: int, status: str, title: str, journal: str, pub_date: str, created: date,
                 completed: date=None, revised: date=None, issue: str=None, pagination: str=None):
        assert pmid > 0, pmid
        assert status in Citation.STATES, repr(status)
        assert title, repr(title)
        assert journal, repr(journal)
        assert pub_date, repr(pub_date)
        assert isinstance(created, date), repr(created)
        assert completed is None or isinstance(completed, date), repr(completed)
        assert revised is None or isinstance(revised, date), repr(revised)
        assert pagination is None or pagination
        assert issue is None or issue
        self.pmid = pmid
        self.status = status
        self.title = title
        self.journal = journal
        self.pub_date = pub_date
        self.issue = issue
        self.pagination = pagination
        self.created = created
        self.completed = completed
        self.revised = revised

    def __str__(self):
        return '{}\n'.format('\t'.join(map(str, [
            NULL(self.pmid), NULL(self.status), STRING(self.title), STRING(self.journal),
            STRING(self.pub_date), NULL(self.issue), NULL(self.pagination),
            DATE(self.created), DATE(self.completed), DATE(self.revised),
            DATE(date.today() if self.modified is None else self.modified)
        ])))

    def __repr__(self):
        return "Citation<{}>".format(self.pmid)

    def __eq__(self, other):
        return isinstance(other, Citation) and \
            self.pmid == other.pmid and \
            self.status == other.status and \
            self.journal == other.journal and \
            self.title == other.title and \
            self.pub_date == other.pub_date and \
            self.issue == other.issue and \
            self.pagination == other.pagination and \
            self.created == other.created and \
            self.completed == other.completed and \
            self.revised == other.revised

    def citation(self):
        issue = '; {}'.format(self.issue) if self.issue else ""
        pagination = ': {}'.format(self.pagination) if self.pagination else ""
        return "{}{}{}".format(self.pub_date, issue, pagination)

    @classmethod
    def insert(cls, data: dict):
        """
        Insert *data* into all relevant tables.
        """
        target_ins = dict(
            (tname, cls.TABLES[tname].insert())
            for tname in cls.TABLENAMES
        )
        conn = _db.engine.connect()
        transaction = conn.begin()

        try:
            if cls.__tablename__ in data and len(data[cls.__tablename__]):
                conn.execute(
                    cls.__table__.insert(), data[cls.__tablename__]
                )

            if Abstract.__tablename__ in data and len(data[Abstract.__tablename__]):
                conn.execute(
                    Abstract.__table__.insert(), data[Abstract.__tablename__]
                )

            if Descriptor.__tablename__ in data and len(data[Descriptor.__tablename__]):
                conn.execute(
                    Descriptor.__table__.insert(), data[Descriptor.__tablename__]
                )

            for tname in cls.TABLENAMES:
                if tname == Descriptor.__tablename__ or tname == Abstract.__tablename__:
                    pass
                elif tname in data and len(data[tname]):
                    conn.execute(target_ins[tname], data[tname])

            transaction.commit()
        except:
            transaction.rollback()
            raise
        finally:
            conn.close()

    @classmethod
    def select(cls, pmids: list, attributes: iter):
        """
        Return the `pmid` and *attributes*
        for each row as a `sqlalchemy.engine.RowProxy`
        that matches one of the *pmids*.
        """
        if not len(pmids):
            return []

        c = cls.__table__.c
        mapping = {col.key: col for col in c}
        columns = [mapping[name] for name in attributes]
        columns.insert(0, c.pmid)
        query = select(columns, c.pmid.in_(pmids))
        return _fetch_all(query)

    @classmethod
    def selectAll(cls, pmids: list):
        """
        Return all columns
        for each row as a `sqlalchemy.engine.RowProxy`
        that matches one of the *pmids*.
        """
        if not len(pmids):
            return []

        c = cls.__table__.c
        query = select([cls.__table__], c.pmid.in_(pmids))
        return _fetch_all(query)

    @classmethod
    def delete(cls, primary_keys: list):
        """
        Delete records and their dependent entities (authors, identifiers,
        etc.) for the given *primary_keys* (a list of PMIDs).
        """
        if not len(primary_keys):
            return

        t = cls.__table__
        query = t.delete(t.c.pmid.in_(primary_keys))
        conn = _db.engine.connect()
        transaction = conn.begin()

        try:
            conn.execute(query)
            transaction.commit()
        except:
            transaction.rollback()
            raise
        finally:
            conn.close()

    @classmethod
    def existing(cls, pmids: list):
        """Return the sub- `set` of all *pmids* that exist in the DB."""
        if not len(pmids):
            return set()

        c = cls.__table__.c
        query = select([c.pmid], c.pmid.in_(pmids))
        conn = _db.engine.connect()

        try:
            return {row[0] for row in conn.execute(query)}
        finally:
            conn.close()

    @classmethod
    def missing(cls, pmids: list):
        """Return the sub- `set` of all *pmids* that do not exist in the DB."""
        return set(pmids) - Citation.existing(pmids)

    @classmethod
    def modifiedBefore(cls, pmids: list, before: date):
        """
        Return the sub- `set` of all *pmids* that have been `modified`
        *before* a `datetime.date` in the DB.
        """
        if not len(pmids):
            return set()

        c = cls.__table__.c
        query = select(
            [c.pmid], c.pmid.in_(pmids) & (c.modified < before)
        )
        conn = _db.engine.connect()

        try:
            return set(row[0] for row in conn.execute(query))
        finally:
            conn.close()



"""
.. py:module:: medic.parser
   :synopsis: An ORM parser for MEDLINE XML records.

.. moduleauthor: Florian Leitner <florian.leitner@gmail.com>
.. License: GNU Affero GPL v3 (http://www.gnu.org/licenses/agpl.html)
"""
import logging
import re
import struct
import types

from xml.etree.ElementTree import iterparse
from datetime import date

from medic.orm import Citation, Abstract, Author, Chemical, Database, \
    Descriptor, Identifier, Keyword, PublicationType, Qualifier, Section

__all__ = ['MedlineXMLParser', 'PubMedXMLParser']

# translate three-letter month strings to integers:
MONTHS_SHORT = (None, 'jan', 'feb', 'mar', 'apr', 'may', 'jun',
                'jul', 'aug', 'sep', 'oct', 'nov', 'dec')

logger = logging.getLogger(__name__)


class State:
    UNDEFINED = 0
    PARSING = 1
    SKIPPING = -1


class Parser:
    """A basic parser implementation for NLM XML citations."""

    def __init__(self, unique=True):
        """
        Create a new parser.

        :param unique: if `True`, citations with VersionID != "1" are skipped
        """
        logger.info('configuring a %sunique %s',
                    "" if unique else "non-", self.__class__.__name__)
        self.unique = unique
        self.events = ('start', 'end') if unique else None
        logger.debug('state: UNDEFINED')
        self._state = State.UNDEFINED
        self.pmid = -1

    def reset(self, pmid):
        self.pmid = pmid

    def skipping(self):
        logger.debug('state: SKIPPING')
        self._state = State.SKIPPING

    def isSkipping(self):
        return State.SKIPPING == self._state

    def undefined(self):
        logger.debug('state: UNDEFINED')
        self._state = State.UNDEFINED

    def isUndefined(self):
        return State.UNDEFINED == self._state

    def parsing(self):
        logger.debug('state: PARSING')
        self._state = State.PARSING

    def isParsing(self):
        return State.PARSING == self._state

    def parse(self, xml_stream):
        try:
            for event, element in iterparse(xml_stream, self.events):
                if event == 'start':
                    self.startElement(element)
                else:
                    for instance in self.yieldInstances(element):
                        yield instance
        except struct.error:
            logger.exception('compressed gzip file is corrupt')

    def startElement(self, element):
        if self.unique and element.tag == 'MedlineCitation':
            version = element.get('VersionID')

            if version is not None and version.strip() != "1":
                logger.debug(
                    'detected a citation with VersionID "%s"',
                    version
                )
                self.skipping()

    def yieldInstances(self, element):
        if element.tag == 'PMID':
            self.PMID(element)
        elif element.tag == 'DeleteCitation':
            for pmid in self.DeleteCitation(element):
                yield pmid
        elif self.isSkipping():
            if element.tag == 'MedlineCitation':
                self.undefined()
        elif hasattr(self, element.tag):
            try:
                for i in self.yieldFromGenerator(element):
                    yield i
            except Exception:
                logger.critical('error while parsing PMID %d', self.pmid)
                raise

    def yieldFromGenerator(self, element):
        logger.debug('processing %s', element.tag)
        instance = getattr(self, element.tag)(element)

        if instance is not None:
            logger.debug('parsed %s', element.tag)

            if isinstance(instance, types.GeneratorType):
                for i in instance:
                    if i is not None:
                        yield i
            else:
                yield instance
        else:
            logger.debug('ignored %s', element.tag)

    @staticmethod
    def DeleteCitation(element):
        for pmid in element.findall('PMID'):
            yield int(pmid.text)

    def MedlineCitation(self, element):
        options = {}

        for name, key in (('DateCompleted', 'completed'),
                          ('DateCreated', 'created'),
                          ('DateRevised', 'revised')):
            e = element.find(name)

            if e is not None:
                options[key] = ParseDate(e)

        created = options['created']
        del options['created']
        status = element.get('Status')
        journal = element.find('MedlineJournalInfo').find(
            'MedlineTA'
        ).text.strip()
        article = element.find('Article')
        title = article.find('ArticleTitle').text

        # ArticleTitle element tags should never be empty when in fact they are sometimes,
        # in violation of the MEDLINE DTD; For example, while PMID 22536004 has a
        # VernacularTitle, its ArticleTitle element is empty.
        if title is None and article.find('VernacularTitle') is not None:
            logging.info('PMID %d had no ArticleTitle; using VernacularTitle instead' % self.pmid)
            title = article.find('VernacularTitle').text

        if title is None:
            logging.warning('could not find any title for PMID %d' % self.pmid)
            title = 'UNKNOWN'  # NB: a DB constraint ensures titles are never empty!

        title = title.strip()
        pub_date = self.parsePubDate(
            article.find('Journal/JournalIssue/PubDate')
        )
        issue = self.parseIssue(article.find('Journal/JournalIssue'))
        pagination = self.parsePagination(
            article.find('Pagination/MedlinePgn')
        )

        if issue:
            options['issue'] = issue
        if pagination:
            options['pagination'] = pagination

        return Citation(self.pmid, status, title, journal,
                        pub_date, created, **options)

    @staticmethod
    def parsePubDate(element):
        medline = element.find('MedlineDate')

        if medline is not None:
            return medline.text.strip()
        else:
            datum = [element.find('Year').text.strip()]

            if element.find('Season') is not None:
                datum.append(element.find('Season').text.strip())
            elif element.find('Month') is not None:
                datum.append(element.find('Month').text.strip())

                if element.find('Day') is not None:
                    datum.append(element.find('Day').text.strip())

            return ' '.join(datum)

    @staticmethod
    def parseIssue(element):
        if element is not None:
            issue = None
            vol = element.find('Volume')
            iss = element.find('Issue')

            if vol is not None and vol.text:
                issue = vol.text.strip()

            if iss is not None and iss.text:
                if issue is None:
                    issue = iss.text.strip()
                else:
                    issue = '{}({})'.format(issue, iss.text.strip())

            return issue

    @staticmethod
    def parsePagination(element):
        if element is not None and element.text:
            return element.text.strip()

    def PMID(self, element):
        pmid = int(element.text)

        if self.isUndefined():
            logger.debug('parsing PMID %i', pmid)
            self.reset(pmid)
            self.parsing()
        elif self.isParsing():
            logger.debug('another PMID %i', pmid)
        elif self.isSkipping():
            logger.info('skipping PMID %i', pmid)
        else:
            logger.error('unknown state %i', self._state)


class MedlineXMLParser(Parser):
    """A parser for (offline) MEDLINE XML (files)."""

    TRUNC_MSG = re.compile(r'\(ABSTRACT TRUNCATED AT \d+ WORDS\)$')

    def __init__(self, *args, **kwargs):
        super(MedlineXMLParser, self).__init__(*args, **kwargs)
        self.seq = 0
        self.namespaces = None

    def reset(self, pmid):
        super(MedlineXMLParser, self).reset(pmid)
        self.namespaces = set()

    def Abstract(self, element):
        for item in self.parseAbstract(element, 'NLM'):
            yield item

    def parseAbstract(self, element, source):
        copy = element.find('CopyrightInformation')

        if copy is not None:
            text = copy.text.strip()
            copy = text if text else None

        # Note: yield Abstract before the Sections!
        yield Abstract(self.pmid, source, copy)
        self.seq = 0

        for child in element.getchildren():
            if child.text is not None and child.text.strip():
                if child.tag == 'AbstractText':
                    yield self.parseAbstractText(child, source)

    def parseAbstractText(self, element, source):
        self.seq += 1
        name = element.get('NlmCategory', 'Unassigned').capitalize()
        content = element.text.strip()
        label = element.get('Label', None)
        truncated = False

        if MedlineXMLParser.TRUNC_MSG.search(content):
            content = MedlineXMLParser.TRUNC_MSG.sub('', content).rstrip()
            truncated = True

            if not content:
                content = ' '  # truncated the whole section; use a space (placeholder)

        return Section(self.pmid, source, self.seq, name, content, label, truncated)

    def AuthorList(self, element):
        for pos, author in enumerate(element.getchildren()):
            yield self.parseAuthor(pos, author)

    def parseAuthor(self, pos, element):
        name, forename, initials, suffix = self.parseAuthorElements(
            element.getchildren()
        )

        if initials == forename and initials is not None:
            # prune the repetition of initials in the forename
            forename = None

        if name is not None:
            return Author(self.pmid, pos + 1, name, initials, forename, suffix)
        else:
            logger.warning(
                'empty or missing Author/LastName or CollectiveName in %i',
                self.pmid
            )
            return None

    def parseAuthorElements(self, children):
        name, forename, initials, suffix = None, None, None, None

        for child in children:
            if child.text is not None:
                text = child.text.strip()
                if child.tag == 'LastName':
                    name = text
                elif child.tag == 'ForeName':
                    forename = text[:256]  # safeguard for size violations
                elif child.tag == 'Initials':
                    initials = text
                elif child.tag == 'Suffix':
                    suffix = text
                elif child.tag == 'CollectiveName':
                    name = text
                    forename = ''
                    initials = ''
                    suffix = ''
                elif child.tag == 'Identifier':
                    pass
                elif child.tag == 'Affiliation':
                    pass
                elif child.tag == 'AffiliationInfo':
                    pass
                else:
                    logger.warning('unknown Author element %s "%s" in %i',
                                   child.tag, text, self.pmid)
            else:
                logger.warning('empty Author element %s in %i"',
                               child.tag, self.pmid)

        if forename and initials:
            if forename.replace(' ', '') == initials:
                forename = None

        return name, forename, initials, suffix

    def ChemicalList(self, element):
        for idx, chemical in enumerate(element.getchildren()):
            e = chemical.find('RegistryNumber')
            uid = None

            if e is not None and e.text is not None and e.text.strip() != "0":
                uid = e.text.strip()

            name = chemical.find('NameOfSubstance')
            yield Chemical(self.pmid, idx + 1, name.text.strip(), uid)

    def DataBank(self, element):
        name = element.find('DataBankName')

        if name is not None and name.text:
            accessions = {acc for acc in map(
                lambda e: e.text.strip() if e.text else None,
                element.find('AccessionNumberList').getchildren()
            ) if acc}

            for acc in accessions:
                yield Database(self.pmid, name.text, acc)

    def ELocationID(self, element):
        ns = element.get('EIdType').strip().lower()

        if ns not in self.namespaces:
            self.namespaces.add(ns)
            return Identifier(self.pmid, ns, element.text.strip())

    def KeywordList(self, element):
        owner = element.get('Owner', 'NLM').strip().upper()
        logger.debug('KeywordList Owner="%s"', owner)

        for cnt, keyword in enumerate(element.getchildren()):
            if keyword.text is not None:
                text = keyword.text.strip()

                if '\r' in text:
                    text = text[:text.find('\r')]

                if text:
                    yield Keyword(
                        self.pmid, owner, cnt + 1, text,
                        keyword.get('MajorTopicYN', 'N') == 'Y',
                    )

    def MedlineCitation(self, element):
        instance = Parser.MedlineCitation(self, element)
        element.clear()
        self.undefined()
        return instance

    def MeshHeadingList(self, element):
        for num, mesh in enumerate(element.getchildren()):
            descriptor = mesh.find('DescriptorName')

            if descriptor is not None and descriptor.text:
                yield self.parseDescriptor(num, descriptor)

            for sub, qualifier in enumerate(mesh.findall('QualifierName')):
                if qualifier.text:
                    yield self.parseQualifier(num, sub, qualifier)

    def parseDescriptor(self, num, element):
        return Descriptor(
            self.pmid, num + 1, element.text.strip(),
            element.get('MajorTopicYN', 'N') == 'Y',
        )

    def parseQualifier(self, num, sub, element):
        return Qualifier(
            self.pmid, num + 1, sub + 1, element.text.strip(),
            element.get('MajorTopicYN', 'N') == 'Y',
        )

    def OtherAbstract(self, element):
        children = element.getchildren()
        source = element.get('Type')

        # parse OtherAbstract only if it is not an abstract that only declares
        # that the publisher has another language version available.
        if (source != 'Publisher' or
            len(children) != 1 or
            children[0].tag != 'AbstractText' or
            children[0].text.strip() !=
                "Abstract available from the publisher."):
            for item in self.parseAbstract(element, source):
                yield item

    def OtherID(self, element):
        if element.get('Source', None) == 'NLM':
            text = element.text.strip()

            if text.startswith('PMC'):
                if 'pmc' not in self.namespaces:
                    self.namespaces.add('pmc')
                    return Identifier(self.pmid, 'pmc', text.split(' ', 1)[0])

    def PublicationType(self, element):
        if element.text:
            return PublicationType(self.pmid, element.text.strip().upper())

    # def VernacularTitle(self, element):
    #     if element.text is not None:
    #         self.seq += 1
    #         return Section(self.pmid, self.seq, 'Vernacular',
    #                        element.text.strip())


class PubMedXMLParser(MedlineXMLParser):
    """A parser for PubMed (eUtils, online) XML."""

    def __init__(self, *args, **kwargs):
        super(PubMedXMLParser, self).__init__(*args, **kwargs)

    def PubmedArticle(self, element):
        element.clear()
        self.undefined()

    def ArticleId(self, element):
        instance = None
        ns = element.get('IdType').strip().lower()
        text = element.text.strip()

        if ns in self.namespaces:
            if re.match('\d[\d\.]+/.+', element.text.strip()) and \
                    'doi' not in self.namespaces:
                self.namespaces.add('doi')
                instance = Identifier(self.pmid, 'doi', text)
            else:
                logger.debug('skipping duplicate %s identifier "%s"', ns, text)
        else:
            self.namespaces.add(ns)
            instance = Identifier(self.pmid, ns, text)

        return instance

    def MedlineCitation(self, element):
        # skip MedlineXMLParser-specific instructions:
        return Parser.MedlineCitation(self, element)


def ParseDate(date_element):
    """Parse a **valid** date that (at least) has to have a Year element."""
    year = int(date_element.find('Year').text)
    month, day = 1, 1
    month_element = date_element.find('Month')
    day_element = date_element.find('Day')

    if month_element is not None:
        month_text = month_element.text.strip()
        try:
            month = int(month_text)
        except ValueError:
            logger.debug('non-numeric Month "%s"', month_text)
            try:
                month = MONTHS_SHORT.index(month_text.lower())
            except ValueError:
                logger.warning('could not parse Month "%s"', month_text)
                month = 1

    if day_element is not None:
        try:
            day = int(day_element.text)
        except (AttributeError, ValueError):
            logger.warning('could not parse Day "%s"', day_element.text)

    return date(year, month, day)



"""
.. py:module:: medic.crud
   :synopsis: I/O CRUD to manage a MEDLINE DB.

.. moduleauthor:: Florian Leitner <florian.leitner@gmail.com>
.. License: GNU Affero GPL v3 (http://www.gnu.org/licenses/agpl.html)
"""
import logging

from functools import partial
from itertools import chain
from gzip import open as gunzip
from os import remove
from os.path import join
from sqlalchemy.exc import IntegrityError, DatabaseError
from sqlalchemy.orm import Session

from medic.orm import Citation, Section, Abstract, Author, Descriptor, \
    Qualifier, Database, Identifier, Chemical, Keyword, PublicationType
from medic.parser import MedlineXMLParser, PubMedXMLParser, Parser
from medic.web import Download
from sqlalchemy.sql import operators

logger = logging.getLogger(__name__)

QUERY_LIMIT = 999
"Maximum number of parameters that can be in an SQLite query; default is 999."


def insert(session: Session, files_or_pmids: iter, uniq: bool) -> bool:
    """Insert all records by parsing the *files* or downloading the *PMIDs*."""
    _add(session, files_or_pmids, lambda i: session.add(i), uniq)


def update(session: Session, files_or_pmids: iter, uniq: bool) -> bool:
    """Update all records in the *files* (paths) or download the *PMIDs*."""
    _add(session, files_or_pmids, lambda i: session.merge(i), uniq)


def select(session: Session, pmids: list([int])) -> iter([Citation]):
    """Return an iterator over all `Citation` records for a list of *PMIDs*."""
    count = 0
    offset = 0
    logger.debug("query limit: %s", QUERY_LIMIT)
    logger.debug("check %s", pmids)

    if pmids:
        while offset < len(pmids):
            for record in session.query(Citation).filter(
                    Citation.pmid.in_(pmids[offset:offset + QUERY_LIMIT])
            ):
                count += 1
                yield record

            offset += QUERY_LIMIT
    else:
        while True:
            start = count

            for record in session.query(Citation).slice(offset, offset + QUERY_LIMIT):
                count += 1
                yield record

            if start == count:
                break

            offset += QUERY_LIMIT

    logger.info("retrieved %i records", count)


def delete(session: Session, pmids: list([int])) -> bool:
    """Delete all records for a list of *PMIDs*."""
    if pmids:
        count = session.query(Citation).filter(Citation.pmid.in_(pmids)).delete(
            synchronize_session=False
        )
    elif isinstance(pmids, list) and len(pmids) == 0:
        count = session.query(Citation).delete(synchronize_session=False)
    else:
        logger.critical("pmids not an [empty] list of integers: %s", repr(pmids))
        return False

    session.commit()
    logger.info("deleted %i records", count)
    return True


def dump(files: iter, output_dir: str, unique: bool, update_all: bool):
    """
    Parse MEDLINE XML files into tabular flat-files for each DB table.

    In addition, a ``delete.txt`` file is generated, containing the PMIDs
    that should first be deleted from the DB before copying the dump.

    :param files: a list of XML files to parse (optionally, gzipped)
    :param output_dir: path to the output directory for the dump
    :param unique: if ``True`` only VersionId == "1" records are dumped
    :param update_all: if ``True`` the PMIDs of all parsed records are
                   added to the list of PMIDs for deletion
    """
    out_stream = {
        Citation.__tablename__:
        open(join(output_dir, "citations.tab"), "wt"),
        Abstract.__tablename__:
        open(join(output_dir, "abstracts.tab"), "wt"),
        Section.__tablename__:
        open(join(output_dir, "sections.tab"), "wt"),
        Descriptor.__tablename__:
        open(join(output_dir, "descriptors.tab"), "wt"),
        Qualifier.__tablename__:
        open(join(output_dir, "qualifiers.tab"), "wt"),
        Author.__tablename__:
        open(join(output_dir, "authors.tab"), "wt"),
        Identifier.__tablename__:
        open(join(output_dir, "identifiers.tab"), "wt"),
        Database.__tablename__:
        open(join(output_dir, "databases.tab"), "wt"),
        PublicationType.__tablename__:
        open(join(output_dir, "publication_types.tab"), "wt"),
        Chemical.__tablename__:
        open(join(output_dir, "chemicals.tab"), "wt"),
        Keyword.__tablename__:
        open(join(output_dir, "keywords.tab"), "wt"),
        'delete': open(join(output_dir, "delete.txt"), "wt"),
    }
    count = 0
    parser = MedlineXMLParser(unique)

    for f in files:
        logger.info('dumping %s', f)

        if f.lower().endswith('.gz'):
            in_stream = gunzip(f, 'rb')
        else:
            in_stream = open(f)

        count += _dump(in_stream, out_stream, parser, update_all)

    for stream in out_stream.values():
        if stream.tell() == 0:
            stream.close()
            remove(join(output_dir, stream.name))
        else:
            stream.close()

    logger.info("parsed %i records", count)


def _dump(in_stream, out_stream: dict, parser: Parser, update_all: bool) -> int:
    count = 0

    for i in parser.parse(in_stream):
        if type(i) == int:
            print(i, file=out_stream['delete'])
        else:
            out_stream[i.__tablename__].write(str(i))

            if i.__tablename__ == Citation.__tablename__:
                count += 1

                if update_all:
                    print(i.pmid, file=out_stream['delete'])

    return count


def _add(session: Session, files_or_pmids: iter, dbHandle, unique: bool=True):
    pmids = []
    count = 0
    initial = session.query(Citation).count() if \
        logger.isEnabledFor(logging.INFO) else 0

    try:
        for arg in files_or_pmids:
            try:
                pmids.append(int(arg))
            except ValueError:
                count += _streamInstances(
                    session, dbHandle, _fromFile(arg, unique)
                )

        if len(pmids):
            count += _downloadAll(session, dbHandle, pmids, unique)

        session.commit()

        if logger.isEnabledFor(logging.INFO):
            final = session.query(Citation).count()
            logger.info('parsed %i entities (records before/after: %i/%i)',
                        count, initial, final)
        return True
    except IntegrityError:
        logger.exception('DB integrity violated (duplicate records?)')
        session.rollback()
        return False
    except DatabaseError:
        logger.exception('adding records failed')
        if session.dirty:
            session.rollback()
        return False


def _streamInstances(session: Session, handle, stream: iter) -> int:
    """
    Stream citations and delete records in DB.

    If PMIDs (integers) are encountered on the stream, they are deleted after
    all citations have been handled.

    :param session: the DB session object (SQL Alchemy)
    :param handle: a DB handle to send ORM instances
    """
    count = 0
    deletion = []

    for citation in _collectCitation(stream):
        if type(citation) == int:
            deletion.append(citation)
        else:
            count += _handleCitation(handle, citation)

    if deletion:
        delete(session, deletion)

    logging.debug("streamed %i citations", count)
    return count


def _collectCitation(stream: iter) -> iter:
    """Collect PMIDs or whole citation lists from the stream."""
    pub_types = []  # ouch - there are non-unique PublicationType entries in PubMed...
    citation = []
    pmid = None

    for instance in stream:
        if type(instance) == int:
            logger.debug("delete PMID %i", instance)
            yield instance
        else:
            if instance.pmid != pmid:
                if citation:
                    yield citation
                    citation = []
                    pub_types = []

                pmid = instance.pmid
                logger.debug("collecting PMID %i", pmid)


            if isinstance(instance, PublicationType):
                if instance.value not in pub_types:
                    citation.append(instance)
                    pub_types.append(instance.value)
            else:
                citation.append(instance)

    if len(citation):
        yield citation


def _handleCitation(handle, instances: list):
    """Handle a list of instances representing a citation."""
    # handle the Citation object first:
    for idx in range(len(instances) - 1, -1, -1):
        if isinstance(instances[idx], Citation):
            logger.debug('handling PMID %s', instances[idx].pmid)
            handle(instances.pop(idx))
            break

    # and everything else after that:
    for i in instances:
        handle(i)

    return 1


def _downloadAll(session: Session, dbHandle,
                 pmids: list, unique: bool=True) -> int:
    """
    Download PubMed XML for a list of PMIDs (integers), parse the streams,
    and send the ORM instances to a DB handle.

    :param session: the SQL Alchemy DB session
    :param dbHandle: a function that takes one instance and sends it to the DB
    :param pmids: the list of PMIDs to download
    :param unique: if ``True``, only VersionID == "1" records are handled.
    """
    parser = PubMedXMLParser(unique)
    pmid_sets = [pmids[100 * i:100 * i + 100]
                 for i in range(len(pmids) // 100 + 1)]
    downloads = map(Download, pmid_sets)
    instances = map(parser.parse, downloads)
    streaming = partial(_streamInstances, session, dbHandle)
    return sum(map(streaming, chain(instances)))


def _fromFile(name: str, unique: bool) -> iter:
    logger.info("parsing %s", name)
    parser = MedlineXMLParser(unique)
    stream = _openFile(name)
    return parser.parse(stream)


def _openFile(name):
    if name.lower().endswith('.gz'):
        # use wrapper to support pre-3.3
        return gunzip(name, 'rb')
    else:
        return open(name)



"""
.. py:module:: medic.web
   :synopsis: Downloading MEDLINE XML records via eUtils.

.. moduleauthor:: Florian Leitner <florian.leitner@gmail.com>
.. License: GNU Affero GPL v3 (http://www.gnu.org/licenses/agpl.html)
"""

import logging

try:
    # noinspection PyCompatibility,PyUnresolvedReferences
    from http.client import HTTPResponse  # only used in function annotations
    # noinspection PyCompatibility,PyUnresolvedReferences
    from urllib.request import build_opener
except ImportError:
    # Python 2.7 compatibility
    # noinspection PyCompatibility,PyUnresolvedReferences
    from httplib import HTTPResponse
    # noinspection PyCompatibility,PyUnresolvedReferences
    from urllib2 import build_opener

from io import StringIO

URL_OPENER = build_opener()
# to avoid having to build a new opener for each request

logger = logging.getLogger(__name__)

EUTILS_URL = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?' \
             'tool=medic&db=pubmed&retmode=xml&rettype=medline&id='
"""
The eUtils URL to which requests for MEDLINE XML records can be made.
Multiple PMIDs may be appended comma-separated to this URL.
"""

DATABANK_LINK = {
    'DOI':
    "http://dx.doi.org/",
    'GDB': "#",
    'GENBANK':
    "http://www.ncbi.nlm.nih.gov/nucgss/",
    'OMIM':
    "http://omim.org/entry/",
    'PDB':
    "http://www.rcsb.org/pdb/explore/explore.do?structureId=",
    'PIR': "#",
    'PMC':
    "http://www.ncbi.nlm.nih.gov/pmc/articles/",
    'RefSeq':
    "http://www.ncbi.nlm.nih.gov/nuccore/",
    'SWISSPROT':
    "http://www.uniprot.org/uniprot/",
    'ClinicalTrials.gov': "#",
    'ISRCTN': "#",
    'GEO': "#",
    'PubChem-Substance':
    "http://pubchem.ncbi.nlm.nih.gov/summary/summary.cgi?sid=",
    'PubChem-Compound':
    "http://pubchem.ncbi.nlm.nih.gov/summary/summary.cgi?cid=",
    'PubChem-BioAssay':
    "http://pubchem.ncbi.nlm.nih.gov/assay/assay.cgi?aid=",
    'PubMed':
    "http://www.ncbi.nlm.nih.gov/pubmed/",
}
"""
Base links to external resources; link target ID needs to be attached to tail.
"""


def Download(pmids: list, timeout: int=60) -> HTTPResponse:
    """
    :param pmids: a list of PMIDs, but no more than `FETCH_SIZE`; values that
        can be cast to string
    :param timeout: seconds to wait for a response

    :return: an open XML stream

    :raises IOError: if the stream from eUtils cannot be opened
    :raises urllib.error.URLError: if the connection to the eUtils URL cannot
        be made
    :raises socket.timeout: if *timeout* seconds have passed before a response
        arrives
    """
    assert len(pmids) <= 100, 'too many PMIDs'
    url = EUTILS_URL + ','.join(map(str, pmids))
    logger.info('fetching %i MEDLINE records from %s', len(pmids), url)
    return URL_OPENER.open(url, timeout=timeout)


def FormatHTML(query):
    file = StringIO()
    p = file.write
    p("""<!doctype html>
<html><head>
  <meta charset="UTF-8"/>
  <title>PubMed Articles</title>
  <script>
function toggle(e) {
    if (e.style.display == 'none')
        e.style.display = 'block'
    else
        e.style.display = 'none'
}
function toggleAll(items) {
    for (var i = items.length; i--; i != 0) {
      toggle(items.item(i))
    }
}
function toggleVisibility(pmid, target) {
    var metadata = document.getElementById(
      pmid
    ).getElementsByTagName('div').item(0)
    var elements = metadata.childNodes;
    for (i = elements.length; i--; i != 0) {
      var e = elements.item(i)
      if (typeof(e.getAttribute) != 'undefined' &&
          e.getAttribute("class") == target) {
        toggle(e)
      }
    }
}
  </script>
</head><body>""")  # .format(file.encoding))

    href = lambda pmid: "{}{}".format(DATABANK_LINK['PubMed'], pmid)
    button = (
        lambda pmid, target, title:
        '<button onclick="toggleVisibility({}, \'{}\')">{}</button>'.format(
            pmid, target, title
        )
    )
    doi = lambda val: '<a href="{}{}">{}</a>'.format(
        DATABANK_LINK['DOI'], val, val
    )
    pmc = (
        lambda val:
        '<a href="{}{}">{}</a>'.format(DATABANK_LINK['PMC'], val, val)
    )
    # dates = lambda rec: 'created: {}{}{}'.format(
    #     rec.created,
    #     "" if rec.completed is None else
    #     ", completed: {}".format(rec.completed),
    #     "" if rec.revised is None else
    #     ", revised: {}".format(rec.revised))
    formatMajor = lambda e: '<b>{}</b>'.format(e.name) if e.major else e.name

    for rec in query:
        logging.debug('writing PMID %i as HTML', rec.pmid)
        citation = '{}, {}'.format(rec.journal, rec.citation())
        link = None
        pt_list = ', '.join(pt.value for pt in rec.publication_types)
        p('<article id={}>'.format(rec.pmid))

        if 'doi' in rec.identifiers:
            link = '<a href="{}{}">'.format(
                DATABANK_LINK['DOI'], rec.identifiers['doi'].value
            )
        elif 'pmc' in rec.identifiers:
            link = '<a href="{}{}">'.format(
                DATABANK_LINK['PMC'], rec.identifiers['pmc'].value
            )

        if link:
            p('<p class="citation"><small>{}{}</a> '
              '({}, PMID:{})</small></p>'.format(
                  link, citation, pt_list, rec.pmid
              ))
        else:
            p('<p class="citation"><small>{} '
              '({}, PMID:{})</small></p>'.format(
                  citation, pt_list, rec.pmid
              ))

        p('  <h1><a href="{}">{}</a></h1>\n  <ol>'.format(
            href(rec.pmid), rec.title
        ))

        for author in rec.authors:
            p('    <li>{}</li>'.format(author.fullName()))

        p('  </ol>')

        for source in rec.abstracts:
            abstract = rec.abstracts[source]
            p('  <h3>{} Abstract</h3>'.format(source))

            for sec in abstract.sections:
                p('  <p title="{}">{}{}</p>'.format(
                    sec.name, "" if sec.label is None else
                    '{}<br/>'.format(sec.label.upper()), sec.content))

            if abstract.copyright:
                p('  <p title="Copyright"><small>{}</small></p>'.format(
                    abstract.copyright
                ))

        p('  <div title="Metadata">')

        if rec.descriptors:
            p('    {}'.format(button(rec.pmid, "mesh", "MeSH Terms")))
        if rec.keywords:
            p('    {}'.format(button(rec.pmid, "kwds", "Keywords")))
        if rec.chemicals:
            p('    {}'.format(button(rec.pmid, "chem", "Chemicals")))
        if rec.databases:
            p('    {}'.format(button(rec.pmid, "xref", "DB Links")))
        if rec.identifiers:
            p('    {}'.format(button(rec.pmid, "ids", "Article IDs")))

        if rec.descriptors:
            p('    <dl class="mesh">')

            for desc in rec.descriptors:
                p('      <dt>{}</dt><dd><ol>'.format(formatMajor(desc)))

                if desc.qualifiers:
                    for qual in desc.qualifiers:
                        p('        <li>{}</li>'.format(formatMajor(qual)))

                p('      </ol></dd>')
            p('    </dl>')

        if rec.chemicals:
            p('    <ul class="chem">')

            for chem in rec.chemicals:
                p('      <li>{}{}</li>'.format(
                    chem.name,
                    "" if chem.uid is None else
                    " ({})".format(chem.uid)
                ))

            p('    </ul>')

        if rec.keywords:
            p('    <dl class="kwds">')
            owner = None

            for kwd in rec.keywords:
                if kwd.owner != owner:
                    if owner is not None:
                        p('        </dd>')

                    p('      <dt>{}</dt><dd>'.format(kwd.owner))
                    owner = kwd.owner

                p('      <li>{}</li>'.format(formatMajor(kwd)))

            p('    </dd></dl>')

        if rec.databases:
            p('    <ul class="xref">')

            for xref in rec.databases:
                try:
                    p('      <li>{} <a href="{}{}">{}</a></li>'.format(
                        xref.name, DATABANK_LINK[xref.name],
                        xref.accession, xref.accession
                    ))
                except KeyError:
                    logging.error('unknown DB name: "{}"'.format(
                        xref.name
                    ))

            p('    </ul>')

        if rec.identifiers:
            p('    <ul class="ids">')

            for ns, i in rec.identifiers.items():
                if ns == 'doi':
                    p('      <li>{}</li>'.format(doi(i.value)))
                elif ns == 'pmc':
                    p('      <li>{}</li>'.format(pmc(i.value)))
                else:
                    p('      <li>{}:{}</li>'.format(ns, i.value))

            p('    </ul>')

        p('  </div>\n</article><hr/>')
    p("""  <script>
window.onload = function() {
    toggleAll(document.getElementsByTagName("ul"))
    toggleAll(document.getElementsByTagName("dl"))
}
  </script>
</body></html>""")
    return file.getvalue()



from distutils.core import setup

try:
    with open('README.rst') as file:
            long_description = file.read()
except IOError:
    long_description = "missing"

setup(
    name='medic',
    version='2.3.0',
    license='GNU GPL v3',
    author='Florian Leitner',
    author_email='florian.leitner@gmail.com',
    url='https://github.com/fnl/medic',
    description='A command line tool to manage a PubMed DB mirror.',
    long_description=long_description,
    install_requires=[
        'sqlalchemy >= 0.8',
    ],
    py_modules=['medic'],
    packages=[
        'medic',
    ],
    package_dir={
        '': 'src',
    },
    scripts=[
        'scripts/medic',
    ],
    data_files=[
        ('share/man/man1', ['medic.1']),
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Software Development :: Libraries',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Linguistic',
    ],
)
